/*

    file:   driver.s
    desc:   Video graphics core driver.

    note:   New video driver for AVGA 2.0, 4cpp VGA-only with keyboard driver (and possibly audio synthesizer).
    	    Screen resolution is no longer configurable and fixed to 168*200@32MHz
            This is a __very__ critical code. Please only modify if you really know what you're doing!

    author: Jaromir Dvorak (md@unicode.cz)

    This file is part of the AVGA platform.
    http://avga.prometheus4.com/


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
*/



#define __ASM__
#define __SFR_OFFSET 0

#include <avr/io.h>
#include "driver.h"


//video pointers
.extern __ramp
.extern _pgmp
.extern _scrp

.extern _compval


//sync&signal generation
.extern c0
.extern wsync
.extern wndd





//keyboard
.extern _kbuf
.extern _kcnt



; register usage
; http://www.nongnu.org/avr-libc/user-manual/FAQ.html#faq_reg_usage
; X: 26 27
; Y: 28 29
; Z: 30 31


;scroll in window
#define scroll		r23

;line in window
#define lineline	r22

;hight of window
#define	wndheight	r15

;line number
#define linecnl	r24
#define linecnh	r25


;
#define zero	r20




;pixel register
#define tmp0     r16

;pixel register 2
#define	tmp2	 r18

#define tmp3    r19

#define	tmp4	r21


;pgm address
#define r_pgm_bitmap_L  r8
#define r_pgm_bitmap_H  r9

;tileset pgm to ram difference
#define pgmramlo r6
#define pgmramhi r7

;compare for high pointer register
#define compval  r17


;16bit pointer to tile
#define ptr1lo  r2
#define ptr1hi  r3

;16bit pointer to next tile
#define ptr2lo  r4
#define ptr2hi  r5


#define	ptrtmp1lo       r12
#define	ptrtmp1hi       r13

#define	ptrtmp2lo	r10
#define	ptrtmp2hi	r11


#define	tmptmp0		r14





;Z: temporary for reading pointers

;X: pointer to pointers






;.global __vectors
;__vectors:

;	push	r0



//interrupt starts one line before first video line scanned

.global	TIMER1_OVF_vect
TIMER1_OVF_vect:

	push	r0
	in	r0, 0x3f	; sreg
	push	r0
	push	r1
	push	r2
	push	r3
	push	r4
	push	r5
	push	r6
	push	r7
	push	r8
	push	r9
	push	r10
	push	r11
	push	r12
	push	r13
	push	r14
	push	r15
	push	r16
	push	r17
	push	r18
	push	r19
	push	r20
	push	r21
	push	r22
	push	r23
	push	r24
	push	r25
	push	r26
	push	r27

	push	r28
	push	r29

	push	r30
	push	r31


	sbi VIDEO_VSYNC_PORT, VIDEO_VSYNC_PIN

	clr zero

synth:	ldi tmp2,1
	out TIFR1, tmp2

	; following code runs at least once every two scan lines. Great time for synthesizer and keyboard driver.
	; start of regular code
	
	


; keyboard driver
lds tmp2, UCSR0A
sbrs tmp2, RXC0
rjmp nokb

lds ZL,   _kbuf
lds ZH,   _kbuf+1
lds tmp3, _kcnt
add ZL, tmp3
adc ZH, zero
lds tmp2, UDR0
st Z, tmp2
inc tmp3
andi tmp3, 7
sts _kcnt, tmp3

nokb:
; end of keyboard driver



;TODO: write synth here...








	; end of regular code

	;are we scanning?
	brie waitln0

	; load line count number
	lds linecnl, c0
        lds linecnh, c0+1

	; last line?
        sbiw linecnl, 1
        breq newframe

	; go to scan?
        cpi linecnl,  0xBA
        ldi tmp0, 0x01
        cpc linecnh,  tmp0     //440
        brne end2
	rjmp scan

	;rjmp end2



newframe:
        ldi linecnl, 0x0D   //525 lines in total
	ldi linecnh, 0x02
       	cbi VIDEO_VSYNC_PORT, VIDEO_VSYNC_PIN

end2:	;store current line number
        sts c0,   linecnl
        sts c0+1, linecnh


	pop	r31
	pop	r30

	pop	r29
	pop	r28

	pop	r27
	pop	r26
	pop	r25
	pop	r24
	pop	r23
	pop	r22
	pop	r21
	pop	r20
	pop	r19
	pop	r18
	pop	r17
	pop	r16
	pop	r15
	pop	r14
	pop	r13
	pop	r12
	pop	r11
	pop	r10
	pop	r9
	pop	r8
	pop	r7
	pop	r6
	pop	r5
	pop	r4
	pop	r3
	pop	r2
	pop	r1
	pop	r0
	out	0x3f, r0	; 63
	pop	r0
	reti







	;;; TODO: sync with timer!!!!
waitln0:in tmp2, TIFR1
	sbrs tmp2,0
	rjmp waitln0
waitln: lds tmp2,TCNT1L   ;wait for visible area
	cpi tmp2,224
	brlo waitln

	; sync modulo 4
	sbrs tmp2,0
	rjmp .
	sbrs tmp2,1
	lpm tmp3, Z

	; sync modulo 8
	;sbrs tmp2,2
	;lpm tmp3, Z
	;sbrs tmp2,1
	;lpm tmp3, Z

	nop
	nop
	rjmp nopreload







        ; if decided to go to render, continue here

scan:   sbiw linecnl,40


	ldi lineline, 0

	
	 ; prepare some constants
         ;lds r_pgm_bitmap_L, _pgmp
         ;lds r_pgm_bitmap_H, _pgmp+1
	 ;lds pgmramlo, __ramp
         ;lds pgmramhi, __ramp+1
         ;sub pgmramlo, r_pgm_bitmap_L
         ;sbc pgmramhi, r_pgm_bitmap_H




         lds YL, wndd
         lds YH, wndd+1

         ldi tmp0, 1
         mov wndheight, tmp0



	; enable "scan break" interrupt
	ldi tmp0, 2
	sts TIMSK1, tmp0

	;clear pending flags
	ser tmp0
	out TIFR1, tmp0


	sei
	;wait for first int to come on
wait:	rjmp wait




endd:   adiw linecnl,40

  	;wsync = 1;
	ldi tmp0,1
	sts wsync, tmp0

  	;__video_timer_select_OVF_IRQ();
	ldi tmp0,1
	sts TIMSK1, tmp0

	;__video_timer_clear_OVF_IRQ();
	ser tmp0
	out TIFR1, tmp0

	rjmp end2


        ;sem jumpne irq...
.global	TIMER1_COMPA_vect       ;TODO: dat na zacatek line.
TIMER1_COMPA_vect:
	out VIDEO_PORT, zero
	;out TIFR1, ff       	; je to zde potreba? neni. flag se cisti automaticky na zacatku nebo az na konci interruptu?    na zacatku.
	pop r1 ;;destroy stack address stored by mcu when executing irq
	pop r1

	sbiw linecnl,1
        breq endd

	sei

	sbrs linecnl,0
	rjmp synth


	;mame 1016 cyklu na linku   @32MHz
	;now compute pointers
	
	
;dorovnani sirky radku. mozna by bylo logictejsi udelat na konci jump na waitln, ale takhle mame o neco cyklu navic.
rjmp .
rjmp .
rjmp .
rjmp .
rjmp .
nop


        dec wndheight
        breq wndload
        inc lineline

        lds r_pgm_bitmap_L, _pgmp
        lds r_pgm_bitmap_H, _pgmp+1

	rjmp .
	rjmp .
	rjmp .
	rjmp .
	rjmp .
	nop

	rjmp ldptr
	
wndload:ld lineline, Y+
	ld wndheight, Y+        ; 9
	ld scroll, Y+
	ld r_pgm_bitmap_L, Y+
        ld r_pgm_bitmap_H, Y+
        sts _pgmp, r_pgm_bitmap_L
        sts _pgmp+1, r_pgm_bitmap_H
        swap scroll
        out VIDEO_DDR, scroll
        swap scroll


ldptr:
	 ;load scrp
         lds XL, _scrp
         lds XH, _scrp+1

	 lds pgmramlo, __ramp            ; __ramp = pointer_v_ram - DRIVER_PGM_BLOCK_COUNT * DRIVER_BLOCK_SIZE;
         lds pgmramhi, __ramp+1
         sub pgmramlo, r_pgm_bitmap_L    ; pgmram = ramp - DRIVER_PGM_BLOCK_COUNT*DRIVER_BLOCK_SIZE - pgmp;
         sbc pgmramhi, r_pgm_bitmap_H

	 ; pricti k scrp DRIVER_COLUMNS + line/8, vysledkem je adresa radku v XH
         mov tmp3, lineline
         lsr tmp3
         lsr tmp3
         lsr tmp3
         ldi tmp2, DRIVER_COLUMNS
         mul tmp3, tmp2
         add XL, r0
         adc XH, r1

	 ; pricti k pointerum ram a pgm (line&7)*4
         mov tmp0, lineline
	 andi tmp0, 7
         lsl tmp0
         lsl tmp0

	 lds compval, _compval
	 ;ldi compval, 29        ;28 means the 28*8 ID is in pgm
         add compval, r_pgm_bitmap_H
         
         
         ;TILESET
         ;
         ; PPPPPPPPP...n...PPPPXXXXXXXXRRRRRRRRR...n...RRR
         ;
         ; P: PGM tiles
         ; R: RAM tiles
         ; X: 8 undefined tiles, depends on PGM tileset base address. If modulo 256 = 0, then all XXXXX are RAM tiles.
         ;


         add r_pgm_bitmap_L, tmp0
         adc r_pgm_bitmap_H, zero   ;porovnava se to prece uz s prictenym.... - r_pgm_bitmap_H se nekdy timto zvetsi o 1
         			    ;ano ale to muze zpusobit glitch...
         			    ;k tomuhle se pricitaji nasobky 32 a pak se to porovnava



	 ; cislo od 0 do 32     - ktera osmice tiles je v ram

	 ; sometimes, it can increase







; a jedeme preloadovat
;;;PRELOAD
;prvni, druha a treti tile se budou preloadovat specialne
;prvni tile: rovnou se vypocita a zustane v Z
;druha tile: preload do ptr1 a po vygenerovani 1. tile dat do Z
;treti tile: preload do ptr2
;dal do RAM:



         ldi tmp3, DRIVER_BLOCK_SIZE	     ; velikost bloku

sc1: 	ld tmp2, X+
	mul tmp2, tmp3
	add r0, r_pgm_bitmap_L
	adc r1, r_pgm_bitmap_H
	movw ZL, r0

	cp ZH, compval ; compare n+2
	brsh ramsc1

pgmsc1: lpm r2, Z+         ;1+12+8+2=23
	lpm r3, Z+
	lpm r4, Z+
	lpm r5, Z+
	rjmp sc2	    ;2

ramsc1: add ZL, pgmramlo
        adc ZH, pgmramhi
	ld r2, Z+         ;2+2+8+8+4 = 24
	ld r3, Z+
	ld r4, Z+
	ld r5, Z+
	rjmp .
	nop

sc2: 	ld tmp2, X+
	mul tmp2, tmp3
	add r0, r_pgm_bitmap_L
	adc r1, r_pgm_bitmap_H
	movw ZL, r0

	cp ZH, compval ; compare n+2
	brsh ramsc2

pgmsc2: lpm tmptmp0, Z+         ;3  +1   = 7
	set                 ;1
	rjmp sc3	    ;2

ramsc2: add ZL, pgmramlo    ;1  +2   = 7
        adc ZH, pgmramhi    ;1
	ld tmptmp0, Z+         ;2
	clt                 ;1

sc3:   	ld tmp2, X+
	mul tmp2, tmp3
	add r0, r_pgm_bitmap_L
	adc r1, r_pgm_bitmap_H
	;movw ptr2lo, r0

	movw ptrtmp1lo, ZL
	movw ptrtmp2lo, r0

	; az zde zacina st

        ;lds ZL, _ptrtmp
        ;lds ZH, _ptrtmp+1
        
        in ZL, SPL
        in ZH, SPH
        sbiw ZL, 45     ;20 tiles + 4 bytes : 44 total

	st Z+,r2
	st Z+,r3
	st Z+,r4
	st Z+,r5

	ld tmp2, X+
	mul tmp2, tmp3
	add r0, r_pgm_bitmap_L
	adc r1, r_pgm_bitmap_H
	st Z+, r0
	st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

	 ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1

         ld tmp2, X+
         mul tmp2, tmp3
         add r0, r_pgm_bitmap_L
         adc r1, r_pgm_bitmap_H
         st Z+, r0
         st Z+, r1






nopreload:

	 ; TODO: STACK!!!


	; scroll & XL:XH

         ;lds XL, _ptrtmp
         ;lds XH, _ptrtmp+1

        in XL, SPL
        in XH, SPH
        sbiw XL, 45     ;20 tiles + 4 bytes : 44 total

         andi scroll, 7
         mov tmp3, scroll
         lsr tmp3
         add XL,tmp3                         ;add scroll/2  to X pointer
         adc XH,zero
         ldi tmp3, 8
         sub tmp3, scroll   ; <-- tohle jde udelat jednou instrukci urcite... xor 0xff ne?


       	; sreg - flag for first tile
 	 in	r0, 0x3f
 	 
 	 

	; ZL:ZH
	movw ZL, ptrtmp1lo
	
	; ptr2lo:ptr2hi
	movw ptr2lo, ptrtmp2lo
	
	;tmp0
	mov tmp0, tmptmp0





; WAIT FOR START OF LINE






;============================================== RENDERING ============================================================

;;;;  co tady musi byt ted - pamatovat si i pres synth ;;;;;;
;v ZL:ZH pointer na tile 1
;v ptr2lo:ptr2hi pointer na tile 2
;v tmp0 prvni pixel tile 1
;v pgmramlo:pgmramhi offset z pgm do ram
;v XL:XH adresa zacatku preloadovanych dat (stack) - pointer se meni podle scrollu!   Kdyby se scroll pocital tady, muzeme to brat ze Stack pointeru...
;v tmp3 pripraveny scroll 0. tile
;v compval hodnota na porovnavani
;a T flag (bit v registru?) jestli prvni tile je v ram nebo pgm



;co z toho prepisuje vlastni rendering:
;ZL:ZH           - rekneme ze to bude jiny registr, udelame mov
;ptr2lo:ptr2hi   - rekneme ze to bude jiny registr, udelame mov
;tmp0            - rekneme ze to bude jiny registr, udelame mov
;XL:XH    	 - tohle muzeme bezpecne vypocitat pri vypoctu scrollu
;tmp3            - tohle muzeme vypocitat pri vypoctu scrollu






;skok muze byt po 1, po 3 a po 5. pixelu. (po 7 ne protoze tam musi bejt swap a movw)
;chceme po co nejvyssim pixelu?
; po 1: je pozde rozhodovat...
; (po 5 ne protoze testovaci intrukce musi bejt mezi 7 a 8. Tam je jediny misto na 1cyklovou ins.
; swapovat se skutecne musi mezi kazdym pixlem...
;OMEZENI pro ram2pgm: ram2pgm musi byt umistene tak aby v ramci renderovani te PGM tile (po te v RAM) to mohlo skocit na pgm2ram
;test na n-1 musi byt pred skokem
;store musi byt bezpodminecne nutne az po testu na pgm,  tedy pokud dalsi tile bude v RAM. jinak si vzdycky prepiseme registry
;add adc musi byt taky v RAM






	 ld tmp2,X+
	 lsr tmp3
         brcs sspeven                        ;can't be 0
         nop
         rjmp sspodd

scloop:	out VIDEO_PORT,tmp2

	ld tmp2, X+
sspodd:	dec tmp3     	       ; mozna bychom mohli dat spis bitovy posun, aby se neprepsal sreg s porovnanim
	out VIDEO_PORT,tmp2

sspeven:swap tmp2
	brne scloop
	movw ptr1lo, ptr2lo   ;n+1 is in ptr1lo:ptr1hi
	out VIDEO_PORT,tmp2    ;8   - posledni pixel, po swapu. A ted co se bude dit dal... dalsi muze byt v pgm nebo v ram


;;;;  co tady musi byt ted: ;;;;;;
;v Z pointer na tile n
;v ptr2 pointer na tile n+1
;v tmp0 prvni pixel. 								v tmp2 druhy pixel? jen pro ram, zatim... idealne tohle ne
;a T flag jestli prvni tile je v ram nebo pgm


	sbrs r0,6
	rjmp ramstr0		;brtc ramstrt0 - bohuzel nejde udelat, ramstrt dale nez 64 words, usetrili bysme jeden cyklus
      	cp ptr2hi, compval
	out VIDEO_PORT,tmp0 			;1

	swap tmp0
	brsh pgm2ram
	nop
pgm1:	out VIDEO_PORT,tmp0			;2

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;3

	swap tmp0
	ld ptr1lo,X+    ; load n+2
	out VIDEO_PORT,tmp0			;4

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;5

	ld ptr1hi,X+    ; load n+2
	swap tmp0
	out VIDEO_PORT,tmp0			;6

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;7

	swap tmp0
	movw ZL, ptr2lo ; fetch n+1			; movw -- musi bezpodminecne nutne byt mezi 7 a 8 pixelem PGM.
	cp ptr1hi, compval ; compare n+2		; TESTOVACI INSTRUKCE -- musi byt bezpodminecne nutne taky mezi 7 a 8
	out VIDEO_PORT,tmp0			;8

; n++ ================================================================================

	lpm tmp0,Z+                   	   	;//a jedeme novou tile (tile1) - ale co kdyz je v ram? - MUSIME MIT 16 pixlu, tedy 2 PGM tily v kodu...
	out VIDEO_PORT,tmp0 			;1

	swap tmp0
	brsh pgm2ram    ; n+1 v ram?
	nop
pgm2:	out VIDEO_PORT,tmp0			;2

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;3

	swap tmp0
	ld ptr2lo,X+        ; load n+2
	out VIDEO_PORT,tmp0			;4

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;5

	ld ptr2hi,X+        ; load n+2
	swap tmp0
	out VIDEO_PORT,tmp0			;6

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;7

	swap tmp0
	movw ZL, ptr1lo	;fetch n+1		; movw -- musi bezpodminecne nutne byt mezi 7 a 8 pixelem PGM.
	cp ptr2hi, compval  ;compare n+2	; TESTOVACI INSTRUKCE -- musi byt bezpodminecne nutne taky mezi 7 a 8
ram2pgm:out VIDEO_PORT,tmp0			;8

; n++ ==================================================================

	lpm tmp0,Z+                   	   	;//a jedeme novou tile (tile1) - ale co kdyz je v ram? - MUSIME MIT 16 pixlu, tedy 2 PGM tily v kodu...
pgmstrt:out VIDEO_PORT,tmp0 			;1

	swap tmp0
	brlo pgm1     ;n+1 v pgm?
	movw ptr1lo, ptr2lo   ;n+1 is in ptr1lo:ptr1hi
pgm2ram:out VIDEO_PORT,tmp0              	;2

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;3

	ld ptr2lo,X+        ; load n+2             ;kdy se komparuje tato tile????
	swap tmp0
	out VIDEO_PORT,tmp0			;4

	lpm tmp0,Z+
	out VIDEO_PORT,tmp0			;5

	ld ptr2hi,X+        ; load n+2
	swap tmp0
	out VIDEO_PORT,tmp0			;6

	lpm tmp0,Z+
ram:	out VIDEO_PORT,tmp0			;7

	swap tmp0
	movw ZL, ptr1lo ;fetch n
        add ZL, pgmramlo
	out VIDEO_PORT,tmp0			;8

; n++ ==================================================================
        adc ZH, pgmramhi
	ld tmp0, Z+
ramstr0:out VIDEO_PORT,tmp0			;1

	ld tmp2, Z+
ramstrt:swap tmp0
	out VIDEO_PORT,tmp0			;2

       	movw ptr1lo, ptr2lo  ; store n+2 as n+1
	ld ptr2lo, X+        ; load n+2
	out VIDEO_PORT,tmp2			;3

	swap tmp2
	ld ptr2hi, X+        ; load n+2
	out VIDEO_PORT,tmp2			;4

	cp ptr1hi, compval      ; compare n+1
	ld tmp2, Z+
	out VIDEO_PORT,tmp2			;5

       	ld tmp0, Z+
	swap tmp2
	out VIDEO_PORT,tmp2			;6

	movw ZL, ptr1lo ;fetch n
	brsh  ram
	cp ptr2hi, compval  ;compare n+2
	out VIDEO_PORT,tmp0			;7

	swap tmp0
	rjmp ram2pgm




















; 32MHz
;cela linka: 1016 cyklu, 812 visible

;1tile ma 4*8:32 cyklu
;22 tiles ma 32*22= 704 visible,  22*10












; cela linka: 953 cyklu  (762 visible)
; double buffering pro 176*200 pixelu:  3*176 = 528 cyklu visible
; zbyva 425 cyklu
; na 23 tiles je potreba 23*32=736 cyklu. Na radek:  368 cyklu. Zbyva 57 cyklu...



; cela linka: 953 cyklu  (762 visible)
; double buffering pro 200*150 pixelu:  3*200 = 600 cyklu visible
; zbyva 353 cyklu
; na 25 tiles je potreba 25*32=800 cyklu. Na radek: 368 cyklu zbyva. 266 cyklu potrebujeme. Pohoda



; @28MHz
; cela linka: 887 cyklu  (711 visible)
; double buffering pro 192*144 pixelu:  3*200 = 600 cyklu visible
; zbyva 287 cyklu
; na 25 tiles je potreba 25*32=800 cyklu. Na radek: 287 cyklu zbyva. 266 cyklu potrebujeme....








